# coding=utf-8
"""Utilities to convert THERM XML files to Material objects."""
from __future__ import division
import xml.etree.ElementTree as ET

from .solid import SolidMaterial
from .cavity import CavityMaterial


def extract_all_materials_from_xml_file(xml_file, gases=None):
    """Extract all Material objects from a THERM XML file.

    Args:
        xml_file: A path to an XML file containing objects Material objects.
        gases: A dictionary with gas names as keys and Gas object instances
            as values. These will be used to reassign the gas that fills
            cavity materials. If None, only solid materials are extracted.

    Returns:
        A tuple with two elements

        -   solid_materials: A list of all SolidMaterial objects in the XML
            file as fairyfly_therm SolidMaterial objects.

        -   cavity_materials: A list of all CavityMaterial objects in the XML
            file as fairyfly_therm CavityMaterial objects.
    """
    # read the file and get the root
    tree = ET.parse(xml_file)
    root = tree.getroot()
    # extract all of the objects
    solid_materials, cavity_materials = [], []
    for mat_obj in root:
        if mat_obj.find('Solid') is not None:
            solid_materials.append(SolidMaterial.from_therm_xml(mat_obj))
            try:
                solid_materials.append(SolidMaterial.from_therm_xml(mat_obj))
            except Exception:  # not a valid solid material
                pass
        elif mat_obj.find('Cavity') is not None and gases is not None:
            try:
                cavity_materials.append(CavityMaterial.from_therm_xml(mat_obj, gases))
            except Exception:  # not a valid cavity material
                pass
    return solid_materials, cavity_materials



# coding=utf-8
"""Gas materials."""
from __future__ import division
import xml.etree.ElementTree as ET

from fairyfly._lockable import lockable
from fairyfly.typing import float_positive, float_in_range, tuple_with_length, \
    uuid_from_therm_id

from ._base import _ResourceObjectBase


@lockable
class PureGas(_ResourceObjectBase):
    """Custom gas gap layer.

    This object allows you to specify specific values for conductivity,
    viscosity and specific heat through the following formula:

    property = A + (B * T) + (C * T ** 2)

    where:

    * A, B, and C = regression coefficients for the gas
    * T = temperature [K]

    Note that setting properties B and C to 0 will mean the property will be
    equal to the A coefficient.

    Args:
        conductivity_coeff_a: First conductivity coefficient.
            Or conductivity in [W/m-K] if b and c coefficients are 0.
        viscosity_coeff_a: First viscosity coefficient.
            Or viscosity in [kg/m-s] if b and c coefficients are 0.
        specific_heat_coeff_a: First specific heat coefficient.
            Or specific heat in [J/kg-K] if b and c coefficients are 0.
        conductivity_coeff_b: Second conductivity coefficient. Default = 0.
        viscosity_coeff_b: Second viscosity coefficient. Default = 0.
        specific_heat_coeff_b: Second specific heat coefficient. Default = 0.
        conductivity_coeff_c: Third conductivity coefficient. Default = 0.
        viscosity_coeff_c: Third viscosity coefficient. Default = 0.
        specific_heat_coeff_c: Third specific heat coefficient. Default = 0.
        specific_heat_ratio: A number for the the ratio of the specific heat at
            constant pressure, to the specific heat at constant volume.
            Default is 1.0 for Air.
        molecular_weight: Number between 20 and 200 for the mass of 1 mol of
            the substance in grams. Default is 20.0.
        identifier: Text string for a unique object ID. Must be a UUID in the
            format xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx. If None, a UUID will
            automatically be generated. (Default: None).

    Properties:
        * identifier
        * display_name
        * conductivity_coeff_a
        * viscosity_coeff_a
        * specific_heat_coeff_a
        * conductivity_coeff_b
        * viscosity_coeff_b
        * specific_heat_coeff_b
        * conductivity_coeff_c
        * viscosity_coeff_c
        * specific_heat_coeff_c
        * specific_heat_ratio
        * molecular_weight
        * protected
        * user_data

    Usage:

    .. code-block:: python

        co2_mat = PureGas(0.0146, 0.000014, 827.73)
        co2_mat.display_name = 'CO2'
        co2_gap.specific_heat_ratio = 1.4
        co2_gap.molecular_weight = 44
        print(co2_gap)
    """
    __slots__ = ('_conductivity_coeff_a', '_viscosity_coeff_a', '_specific_heat_coeff_a',
                 '_conductivity_coeff_b', '_viscosity_coeff_b', '_specific_heat_coeff_b',
                 '_conductivity_coeff_c', '_viscosity_coeff_c', '_specific_heat_coeff_c',
                 '_specific_heat_ratio', '_molecular_weight')

    def __init__(
            self, conductivity_coeff_a, viscosity_coeff_a, specific_heat_coeff_a,
            conductivity_coeff_b=0, viscosity_coeff_b=0, specific_heat_coeff_b=0,
            conductivity_coeff_c=0, viscosity_coeff_c=0, specific_heat_coeff_c=0,
            specific_heat_ratio=1.0, molecular_weight=20.0, identifier=None):
        """Initialize custom gas energy material."""
        _ResourceObjectBase.__init__(self, identifier)
        self.conductivity_coeff_a = conductivity_coeff_a
        self.viscosity_coeff_a = viscosity_coeff_a
        self.specific_heat_coeff_a = specific_heat_coeff_a
        self.conductivity_coeff_b = conductivity_coeff_b
        self.viscosity_coeff_b = viscosity_coeff_b
        self.specific_heat_coeff_b = specific_heat_coeff_b
        self.conductivity_coeff_c = conductivity_coeff_c
        self.viscosity_coeff_c = viscosity_coeff_c
        self.specific_heat_coeff_c = specific_heat_coeff_c
        self.specific_heat_ratio = specific_heat_ratio
        self.molecular_weight = molecular_weight

    @property
    def conductivity_coeff_a(self):
        """Get or set the first conductivity coefficient."""
        return self._conductivity_coeff_a

    @conductivity_coeff_a.setter
    def conductivity_coeff_a(self, coeff):
        self._conductivity_coeff_a = float(coeff)

    @property
    def viscosity_coeff_a(self):
        """Get or set the first viscosity coefficient."""
        return self._viscosity_coeff_a

    @viscosity_coeff_a.setter
    def viscosity_coeff_a(self, coeff):
        self._viscosity_coeff_a = float_positive(coeff)

    @property
    def specific_heat_coeff_a(self):
        """Get or set the first specific heat coefficient."""
        return self._specific_heat_coeff_a

    @specific_heat_coeff_a.setter
    def specific_heat_coeff_a(self, coeff):
        self._specific_heat_coeff_a = float_positive(coeff)

    @property
    def conductivity_coeff_b(self):
        """Get or set the second conductivity coefficient."""
        return self._conductivity_coeff_b

    @conductivity_coeff_b.setter
    def conductivity_coeff_b(self, coeff):
        self._conductivity_coeff_b = float(coeff)

    @property
    def viscosity_coeff_b(self):
        """Get or set the second viscosity coefficient."""
        return self._viscosity_coeff_b

    @viscosity_coeff_b.setter
    def viscosity_coeff_b(self, coeff):
        self._viscosity_coeff_b = float(coeff)

    @property
    def specific_heat_coeff_b(self):
        """Get or set the second specific heat coefficient."""
        return self._specific_heat_coeff_b

    @specific_heat_coeff_b.setter
    def specific_heat_coeff_b(self, coeff):
        self._specific_heat_coeff_b = float(coeff)

    @property
    def conductivity_coeff_c(self):
        """Get or set the third conductivity coefficient."""
        return self._conductivity_coeff_c

    @conductivity_coeff_c.setter
    def conductivity_coeff_c(self, coeff):
        self._conductivity_coeff_c = float(coeff)

    @property
    def viscosity_coeff_c(self):
        """Get or set the third viscosity coefficient."""
        return self._viscosity_coeff_c

    @viscosity_coeff_c.setter
    def viscosity_coeff_c(self, coeff):
        self._viscosity_coeff_c = float(coeff)

    @property
    def specific_heat_coeff_c(self):
        """Get or set the third specific heat coefficient."""
        return self._specific_heat_coeff_c

    @specific_heat_coeff_c.setter
    def specific_heat_coeff_c(self, coeff):
        self._specific_heat_coeff_c = float(coeff)

    @property
    def specific_heat_ratio(self):
        """Get or set the specific heat ratio."""
        return self._specific_heat_ratio

    @specific_heat_ratio.setter
    def specific_heat_ratio(self, number):
        number = float(number)
        assert 1 <= number, 'Input specific_heat_ratio ({}) must be > 1.'.format(number)
        self._specific_heat_ratio = number

    @property
    def molecular_weight(self):
        """Get or set the molecular weight."""
        return self._molecular_weight

    @molecular_weight.setter
    def molecular_weight(self, number):
        self._molecular_weight = float_in_range(
            number, 2.0, 300.0, 'gas material molecular weight')

    @property
    def conductivity(self):
        """Conductivity of the gas in the absence of convection at 0C [W/m-K]."""
        return self.conductivity_at_temperature(273.15)

    @property
    def viscosity(self):
        """Viscosity of the gas at 0C [kg/m-s]."""
        return self.viscosity_at_temperature(273.15)

    @property
    def specific_heat(self):
        """Specific heat of the gas at 0C [J/kg-K]."""
        return self.specific_heat_at_temperature(273.15)

    @property
    def density(self):
        """Density of the gas at 0C and sea-level pressure [J/kg-K]."""
        return self.density_at_temperature(273.15)

    @property
    def prandtl(self):
        """Prandtl number of the gas at 0C."""
        return self.prandtl_at_temperature(273.15)

    def conductivity_at_temperature(self, t_kelvin):
        """Get the conductivity of the gas [W/m-K] at a given Kelvin temperature."""
        return self.conductivity_coeff_a + self.conductivity_coeff_b * t_kelvin + \
            self.conductivity_coeff_c * t_kelvin ** 2

    def viscosity_at_temperature(self, t_kelvin):
        """Get the viscosity of the gas [kg/m-s] at a given Kelvin temperature."""
        return self.viscosity_coeff_a + self.viscosity_coeff_b * t_kelvin + \
            self.viscosity_coeff_c * t_kelvin ** 2

    def specific_heat_at_temperature(self, t_kelvin):
        """Get the specific heat of the gas [J/kg-K] at a given Kelvin temperature."""
        return self.specific_heat_coeff_a + self.specific_heat_coeff_b * t_kelvin + \
            self.specific_heat_coeff_c * t_kelvin ** 2

    def density_at_temperature(self, t_kelvin, pressure=101325):
        """Get the density of the gas [kg/m3] at a given temperature and pressure.

        This method uses the ideal gas law to estimate the density.

        Args:
            t_kelvin: The average temperature of the gas cavity in Kelvin.
            pressure: The average pressure of the gas cavity in Pa.
                Default is 101325 Pa for standard pressure at sea level.
        """
        return (pressure * self.molecular_weight * 0.001) / (8.314 * t_kelvin)

    def prandtl_at_temperature(self, t_kelvin):
        """Get the Prandtl number of the gas at a given Kelvin temperature."""
        return self.viscosity_at_temperature(t_kelvin) * \
            self.specific_heat_at_temperature(t_kelvin) / \
            self.conductivity_at_temperature(t_kelvin)

    @classmethod
    def from_therm_xml(cls, xml_element):
        """Create PureGas from an XML element of a THERM PureGas material.

        Args:
            xml_element: An XML element of a THERM PureGas material.
        """
        # get the identifier, molecular weight and specific heat ratio
        xml_uuid = xml_element.find('UUID')
        identifier = xml_uuid.text
        if len(identifier) == 31:
            identifier = uuid_from_therm_id(identifier)
        xml_prop = xml_element.find('Properties')
        xml_mw = xml_prop.find('MolecularWeight')
        molecular_weight = xml_mw.text
        xml_shr = xml_prop.find('SpecificHeatRatio')
        specific_heat_ratio = xml_shr.text
        # extract the conductivity curve
        xml_cond = xml_prop.find('Conductivity')
        xml_c_a = xml_cond.find('A')
        conductivity_coeff_a = xml_c_a.text
        xml_c_b = xml_cond.find('B')
        conductivity_coeff_b = xml_c_b.text
        xml_c_c = xml_cond.find('C')
        conductivity_coeff_c = xml_c_c.text
        # extract the viscosity curve
        xml_vis = xml_prop.find('Viscosity')
        xml_v_a = xml_vis.find('A')
        viscosity_coeff_a = xml_v_a.text
        xml_v_b = xml_vis.find('B')
        viscosity_coeff_b = xml_v_b.text
        xml_v_c = xml_vis.find('C')
        viscosity_coeff_c = xml_v_c.text
        # extract the specific heat curve
        xml_sh = xml_prop.find('SpecificHeat')
        xml_sh_a = xml_sh.find('A')
        specific_heat_coeff_a = xml_sh_a.text
        xml_sh_b = xml_sh.find('B')
        specific_heat_coeff_b = xml_sh_b.text
        xml_sh_c = xml_sh.find('C')
        specific_heat_coeff_c = xml_sh_c.text
        # create the PureGas material
        mat = PureGas(
            conductivity_coeff_a, viscosity_coeff_a, specific_heat_coeff_a,
            conductivity_coeff_b, viscosity_coeff_b, specific_heat_coeff_b,
            conductivity_coeff_c, viscosity_coeff_c, specific_heat_coeff_c,
            specific_heat_ratio, molecular_weight, identifier=identifier)
        # assign the name if it is specified
        xml_name = xml_element.find('Name')
        if xml_name is not None:
            mat.display_name = xml_name.text
        xml_protect = xml_element.find('Protected')
        if xml_protect is not None:
            mat.protected = True if xml_protect.text == 'true' else False
        return mat

    @classmethod
    def from_therm_xml_str(cls, xml_str):
        """Create a PureGas from an XML text string of a THERM PureGas.

        Args:
            xml_str: An XML text string of a THERM PureGas.
        """
        root = ET.fromstring(xml_str)
        return cls.from_therm_xml(root)

    @classmethod
    def from_dict(cls, data):
        """Create a PureGas from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "type": 'PureGas',
            "identifier": '7b4a5a47-ebec-4d95-b028-a78485130c34',
            "display_name": 'CO2'
            "conductivity_coeff_a": 0.0146,
            "viscosity_coeff_a": 0.000014,
            "specific_heat_coeff_a": 827.73,
            "specific_heat_ratio": 1.4
            "molecular_weight": 44
            }
        """
        assert data['type'] == 'PureGas', \
            'Expected PureGas. Got {}.'.format(data['type'])
        con_b = 0 if 'conductivity_coeff_b' not in data else data['conductivity_coeff_b']
        vis_b = 0 if 'viscosity_coeff_b' not in data else data['viscosity_coeff_b']
        sph_b = 0 if 'specific_heat_coeff_b' not in data \
            else data['specific_heat_coeff_b']
        con_c = 0 if 'conductivity_coeff_c' not in data else data['conductivity_coeff_c']
        vis_c = 0 if 'viscosity_coeff_c' not in data else data['viscosity_coeff_c']
        sph_c = 0 if 'specific_heat_coeff_c' not in data \
            else data['specific_heat_coeff_c']
        sphr = 1.0 if 'specific_heat_ratio' not in data else data['specific_heat_ratio']
        mw = 20.0 if 'molecular_weight' not in data else data['molecular_weight']
        new_obj = cls(
            data['conductivity_coeff_a'], data['viscosity_coeff_a'],
            data['specific_heat_coeff_a'],
            con_b, vis_b, sph_b, con_c, vis_c, sph_c, sphr, mw,
            identifier=data['identifier'])
        if 'display_name' in data and data['display_name'] is not None:
            new_obj.display_name = data['display_name']
        if 'protected' in data and data['protected'] is not None:
            new_obj.protected = data['protected']
        if 'user_data' in data and data['user_data'] is not None:
            new_obj.user_data = data['user_data']
        return new_obj

    def to_therm_xml(self, gases_element=None):
        """Get an THERM XML element of the gas.

        Args:
            gases_element: An optional XML Element for the Gases to which the
                generated objects will be added. If None, a new XML Element
                will be generated.

        .. code-block:: xml

            <PureGas>
                <UUID>8d33196f-f052-46e6-8353-bccb9a779f9c</UUID>
                <Name>Air</Name>
                <Protected>true</Protected>
                <Properties>
                    <MolecularWeight>28.97</MolecularWeight>
                    <SpecificHeatRatio>1.4</SpecificHeatRatio>
                    <Conductivity>
                        <A>0.002873</A>
                        <B>7.76e-05</B>
                        <C>0</C>
                    </Conductivity>
                    <Viscosity>
                        <A>3.723e-06</A>
                        <B>4.94e-08</B>
                        <C>0</C>
                    </Viscosity>
                    <SpecificHeat>
                        <A>1002.737</A>
                        <B>0.012324</B>
                        <C>0</C>
                    </SpecificHeat>
                </Properties>
            </PureGas>
        """
        # create a new Materials element if one is not specified
        if gases_element is not None:
            xml_mat = ET.SubElement(gases_element, 'PureGas')
        else:
            xml_mat = ET.Element('PureGas')
        # add all of the required basic attributes
        xml_id = ET.SubElement(xml_mat, 'UUID')
        xml_id.text = self.identifier
        xml_name = ET.SubElement(xml_mat, 'Name')
        xml_name.text = self.display_name
        xml_protect = ET.SubElement(xml_mat, 'Protected')
        xml_protect.text = 'true' if self.protected else 'false'
        xml_prop = ET.SubElement(xml_mat, 'Properties')
        # molecular weight and specific heat ratio
        xml_mw = ET.SubElement(xml_prop, 'MolecularWeight')
        xml_mw.text = str(self.molecular_weight)
        xml_shr = ET.SubElement(xml_prop, 'SpecificHeatRatio')
        xml_shr.text = str(self.specific_heat_ratio)
        # add the conductivity curve
        xml_cond = ET.SubElement(xml_prop, 'Conductivity')
        xml_cond_a = ET.SubElement(xml_cond, 'A')
        xml_cond_a.text = str(self.conductivity_coeff_a)
        xml_cond_b = ET.SubElement(xml_cond, 'B')
        xml_cond_b.text = str(self.conductivity_coeff_b)
        xml_cond_c = ET.SubElement(xml_cond, 'C')
        xml_cond_c.text = str(self.conductivity_coeff_c)
        # add the viscosity curve
        xml_vis = ET.SubElement(xml_prop, 'Viscosity')
        xml_vis_a = ET.SubElement(xml_vis, 'A')
        xml_vis_a.text = str(self.viscosity_coeff_a)
        xml_vis_b = ET.SubElement(xml_vis, 'B')
        xml_vis_b.text = str(self.viscosity_coeff_b)
        xml_vis_c = ET.SubElement(xml_vis, 'C')
        xml_vis_c.text = str(self.viscosity_coeff_c)
        # add the specific heat curve
        xml_sh = ET.SubElement(xml_prop, 'SpecificHeat')
        xml_sh_a = ET.SubElement(xml_sh, 'A')
        xml_sh_a.text = str(self.specific_heat_coeff_a)
        xml_sh_b = ET.SubElement(xml_sh, 'B')
        xml_sh_b.text = str(self.specific_heat_coeff_b)
        xml_sh_c = ET.SubElement(xml_sh, 'C')
        xml_sh_c.text = str(self.specific_heat_coeff_c)
        return xml_mat

    def to_therm_xml_str(self):
        """Get an THERM XML string of the gas."""
        xml_root = self.to_therm_xml()
        try:  # try to indent the XML to make it read-able
            ET.indent(xml_root)
            return ET.tostring(xml_root, encoding='unicode')
        except AttributeError:  # we are in Python 2 and no indent is available
            return ET.tostring(xml_root)

    def to_dict(self):
        """PureGas dictionary representation."""
        base = {
            'type': 'PureGas',
            'identifier': self.identifier,
            'conductivity_coeff_a': self.conductivity_coeff_a,
            'viscosity_coeff_a': self.viscosity_coeff_a,
            'specific_heat_coeff_a': self.specific_heat_coeff_a,
            'conductivity_coeff_b': self.conductivity_coeff_b,
            'viscosity_coeff_b': self.viscosity_coeff_b,
            'specific_heat_coeff_b': self.specific_heat_coeff_b,
            'conductivity_coeff_c': self.conductivity_coeff_c,
            'viscosity_coeff_c': self.viscosity_coeff_c,
            'specific_heat_coeff_c': self.specific_heat_coeff_c,
            'specific_heat_ratio': self.specific_heat_ratio,
            'molecular_weight': self.molecular_weight
        }
        if self._display_name is not None:
            base['display_name'] = self.display_name
        base['protected'] = self._protected
        if self._user_data is not None:
            base['user_data'] = self.user_data
        return base

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (self.identifier, self.conductivity_coeff_a,
                self.viscosity_coeff_a, self.specific_heat_coeff_a,
                self.conductivity_coeff_b, self.viscosity_coeff_b,
                self.specific_heat_coeff_b, self.conductivity_coeff_c,
                self.viscosity_coeff_c, self.specific_heat_coeff_c,
                self.specific_heat_ratio, self.molecular_weight)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, PureGas) and \
            self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return 'THERM Pure Gas: {}'.format(self.display_name)

    def __copy__(self):
        new_obj = PureGas(
            self.conductivity_coeff_a, self.viscosity_coeff_a, self.specific_heat_coeff_a,
            self.conductivity_coeff_b, self.viscosity_coeff_b, self.specific_heat_coeff_b,
            self.conductivity_coeff_c, self.viscosity_coeff_c, self.specific_heat_coeff_c,
            self.specific_heat_ratio, self.molecular_weight, identifier=self.identifier)
        new_obj._display_name = self._display_name
        new_obj._protected = self._protected
        new_obj._user_data = None if self._user_data is None else self._user_data.copy()
        return new_obj


@lockable
class Gas(_ResourceObjectBase):
    """Gas gap material defined by a mixture of gases.

    Args:
        pure_gases: A list of PureGas objects describing the types of gas in the gap.
        gas_fractions: A list of fractional numbers describing the volumetric
            fractions of gas types in the mixture.  This list must align with
            the pure_gases input list and must sum to 1.
        identifier: Text string for a unique object ID. Must be a UUID in the
            format xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx. If None, a UUID will
            automatically be generated. (Default: None).

    Properties:
        * identifier
        * display_name
        * pure_gases
        * gas_fractions
        * gas_count
        * protected
        * user_data
    """
    __slots__ = ('_gas_count', '_pure_gases', '_gas_fractions')

    def __init__(self, pure_gases, gas_fractions, identifier=None):
        """Initialize gas mixture material."""
        _ResourceObjectBase.__init__(self, identifier)
        try:  # check the number of gases
            self._gas_count = len(pure_gases)
        except (TypeError, ValueError):
            raise TypeError(
                'Expected list for pure_gases. Got {}.'.format(type(pure_gases)))
        assert 1 <= self._gas_count, 'Number of gases in gas mixture must be ' \
            'greater than 1. Got {}.'.format(self._gas_count)
        self.pure_gases = pure_gases
        self.gas_fractions = gas_fractions

    @property
    def pure_gases(self):
        """Get or set a tuple of text describing the gases in the gas gap layer."""
        return self._pure_gases

    @pure_gases.setter
    def pure_gases(self, value):
        try:
            if not isinstance(value, tuple):
                value = tuple(value)
        except TypeError:
            raise TypeError('Expected list or tuple for pure_gases. '
                            'Got {}'.format(type(value)))
        for g in value:
            assert isinstance(g, PureGas), 'Expected PureGas' \
                ' material for Gas. Got {}.'.format(type(g))
        assert len(value) > 0, 'Gas must possess at least one pure gas.'
        self._pure_gases = value

    @property
    def gas_fractions(self):
        """Get or set a tuple of numbers the fractions of gases in the gas gap layer."""
        return self._gas_fractions

    @gas_fractions.setter
    def gas_fractions(self, g_fracs):
        self._gas_fractions = tuple_with_length(
            g_fracs, self._gas_count, float, 'gas mixture gas_fractions')
        assert sum(self._gas_fractions) == 1, 'Gas fractions must sum to 1. ' \
            'Got {}.'.format(sum(self._gas_fractions))

    @property
    def molecular_weight(self):
        """Get the gas molecular weight."""
        return sum(tuple(gas.molecular_weight * frac for gas, frac
                         in zip(self._pure_gases, self._gas_fractions)))

    @property
    def gas_count(self):
        """An integer indicating the number of gases in the mixture."""
        return self._gas_count

    @property
    def conductivity(self):
        """Conductivity of the gas in the absence of convection at 0C [W/m-K]."""
        return self.conductivity_at_temperature(273.15)

    @property
    def viscosity(self):
        """Viscosity of the gas at 0C [kg/m-s]."""
        return self.viscosity_at_temperature(273.15)

    @property
    def specific_heat(self):
        """Specific heat of the gas at 0C [J/kg-K]."""
        return self.specific_heat_at_temperature(273.15)

    @property
    def density(self):
        """Density of the gas at 0C and sea-level pressure [J/kg-K]."""
        return self.density_at_temperature(273.15)

    @property
    def prandtl(self):
        """Prandtl number of the gas at 0C."""
        return self.prandtl_at_temperature(273.15)

    def conductivity_at_temperature(self, t_kelvin):
        """Get the conductivity of the gas [W/m-K] at a given Kelvin temperature."""
        return self._weighted_avg_coeff_property('conductivity_coeff', t_kelvin)

    def viscosity_at_temperature(self, t_kelvin):
        """Get the viscosity of the gas [kg/m-s] at a given Kelvin temperature."""
        return self._weighted_avg_coeff_property('viscosity_coeff', t_kelvin)

    def specific_heat_at_temperature(self, t_kelvin):
        """Get the specific heat of the gas [J/kg-K] at a given Kelvin temperature."""
        return self._weighted_avg_coeff_property('specific_heat_coeff', t_kelvin)

    def density_at_temperature(self, t_kelvin, pressure=101325):
        """Get the density of the gas [kg/m3] at a given temperature and pressure.

        This method uses the ideal gas law to estimate the density.

        Args:
            t_kelvin: The average temperature of the gas cavity in Kelvin.
            pressure: The average pressure of the gas cavity in Pa.
                Default is 101325 Pa for standard pressure at sea level.
        """
        return (pressure * self.molecular_weight * 0.001) / (8.314 * t_kelvin)

    def prandtl_at_temperature(self, t_kelvin):
        """Get the Prandtl number of the gas at a given Kelvin temperature."""
        return self.viscosity_at_temperature(t_kelvin) * \
            self.specific_heat_at_temperature(t_kelvin) / \
            self.conductivity_at_temperature(t_kelvin)

    @classmethod
    def from_therm_xml(cls, xml_element, pure_gases):
        """Create Gas from an XML element of a THERM Gas material.

        Args:
            xml_element: An XML element of a THERM Gas material.
            pure_gases: A dictionary with pure gas names as keys and PureGas
                object instances as values. These will be used to reassign
                the pure gases that make up this gas.
        """
        # get the identifier, gases and initialize the object
        xml_uuid = xml_element.find('UUID')
        identifier = xml_uuid.text
        if len(identifier) == 31:
            identifier = uuid_from_therm_id(identifier)
        xml_comps = xml_element.find('Components')
        pure_gas_objs, gas_fractions = [], []
        for xml_comp in xml_comps:
            xml_gas_name = xml_comp.find('PureGas')
            try:
                pure_gas_objs.append(pure_gases[xml_gas_name.text])
            except KeyError as e:
                raise ValueError('Failed to find {} in pure gases.'.format(e))
            xml_gas_fract = xml_comp.find('Fraction')
            gas_fractions.append(xml_gas_fract.text)
        mat = Gas(pure_gas_objs, gas_fractions, identifier=identifier)
        # assign the name if it is specified
        xml_name = xml_element.find('Name')
        if xml_name is not None:
            mat.display_name = xml_name.text
        xml_protect = xml_element.find('Protected')
        if xml_protect is not None:
            mat.protected = True if xml_protect.text == 'true' else False
        return mat

    @classmethod
    def from_therm_xml_str(cls, xml_str, pure_gases):
        """Create a Gas from an XML text string of a THERM Gas.

        Args:
            xml_str: An XML text string of a THERM Gas.
            pure_gases: A dictionary with pure gas names as keys and PureGas
                object instances as values. These will be used to reassign
                the pure gases that make up this gas gap.
        """
        root = ET.fromstring(xml_str)
        return cls.from_therm_xml(root, pure_gases)

    @classmethod
    def from_dict(cls, data):
        """Create a Gas from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            'type': 'Gas',
            'identifier': 'e34f4b07-a012-4142-ae29-d7967c921c71',
            'display_name': 'Argon Mixture',
            'pure_gases': [{}, {}],  # list of PureGas objects
            'gas_fractions': [0.95, 0.05]
            }
        """
        assert data['type'] == 'Gas', 'Expected Gas. Got {}.'.format(data['type'])
        pure_gases = [PureGas.from_dict(pg) for pg in data['pure_gases']]
        new_obj = cls(pure_gases, data['gas_fractions'], data['identifier'])
        cls._assign_optional_from_dict(new_obj, data)
        return new_obj

    @classmethod
    def from_dict_abridged(cls, data, pure_gases):
        """Create a Gas from an abridged dictionary.

        Args:
            data: An GasAbridged dictionary.
            pure_gases: A dictionary with pure gas identifiers as keys and PureGas
                object instances as values. These will be used to reassign
                the pure gases that make up this gas gap.

        .. code-block:: python

            {
            'type': 'GasAbridged',
            'identifier': 'e34f4b07-a012-4142-ae29-d7967c921c71',
            'display_name': 'Argon Mixture',
            'pure_gases': [
                'ca280a4b-aba9-416f-9443-484285d52227',
                'ba65b928-f766-4044-bc17-e53c42040bde'
            ],
            'gas_fractions': [0.95, 0.05]
            }
        """
        assert data['type'] == 'GasAbridged', \
            'Expected GasAbridged. Got {}.'.format(data['type'])
        try:
            pure_gas_objs = [pure_gases[mat_id] for mat_id in data['pure_gases']]
        except KeyError as e:
            raise ValueError('Failed to find {} in pure gases.'.format(e))
        new_obj = cls(pure_gas_objs, data['gas_fractions'], data['identifier'])
        cls._assign_optional_from_dict(new_obj, data)
        return new_obj

    @staticmethod
    def _assign_optional_from_dict(new_obj, data):
        """Assign optional attributes when serializing from dict."""
        if 'display_name' in data and data['display_name'] is not None:
            new_obj.display_name = data['display_name']
        if 'protected' in data and data['protected'] is not None:
            new_obj.protected = data['protected']
        if 'user_data' in data and data['user_data'] is not None:
            new_obj.user_data = data['user_data']

    def to_therm_xml(self, gases_element=None):
        """Get an THERM XML element of the gas.

        Note that this method only outputs a single element for the gas and,
        to write the full gas into an XML, the gas's pure gases
        must also be written.

        Args:
            gases_element: An optional XML Element for the Gases to which the
                generated objects will be added. If None, a new XML Element
                will be generated.

        .. code-block:: xml

            <Gas>
                <UUID>6c2409e9-5296-46c1-be11-9029b59a549b</UUID>
                <Name>Air</Name>
                <Protected>true</Protected>
                <Components>
                    <Component>
                        <Fraction>1</Fraction>
                        <PureGas>Air</PureGas>
                    </Component>
                </Components>
            </Gas>
        """
        # create a new Materials element if one is not specified
        if gases_element is not None:
            xml_mat = ET.SubElement(gases_element, 'Gas')
        else:
            xml_mat = ET.Element('Gas')
        # add all of the required basic attributes
        xml_id = ET.SubElement(xml_mat, 'UUID')
        xml_id.text = self.identifier
        xml_name = ET.SubElement(xml_mat, 'Name')
        xml_name.text = self.display_name
        xml_protect = ET.SubElement(xml_mat, 'Protected')
        xml_protect.text = 'true' if self.protected else 'false'
        # add the gas components
        xml_comps = ET.SubElement(xml_mat, 'Components')
        for pure_gas, gas_fract in zip(self.pure_gases, self.gas_fractions):
            xml_comp = ET.SubElement(xml_comps, 'Component')
            xml_fact = ET.SubElement(xml_comp, 'Fraction')
            xml_fact.text = str(gas_fract)
            xml_gas = ET.SubElement(xml_comp, 'PureGas')
            xml_gas.text = pure_gas.display_name
        return xml_mat

    def to_therm_xml_str(self):
        """Get an THERM XML string of the gas."""
        xml_root = self.to_therm_xml()
        try:  # try to indent the XML to make it read-able
            ET.indent(xml_root)
            return ET.tostring(xml_root, encoding='unicode')
        except AttributeError:  # we are in Python 2 and no indent is available
            return ET.tostring(xml_root)

    def to_dict(self, abridged=False):
        """Gas dictionary representation."""
        base = {'type': 'Gas'} if not abridged else {'type': 'GasAbridged'}
        base['identifier'] = self.identifier
        base['pure_gases'] = \
            [m.identifier for m in self.pure_gases] if abridged else \
            [m.to_dict() for m in self.pure_gases]
        base['gas_fractions'] = self.gas_fractions
        if self._display_name is not None:
            base['display_name'] = self.display_name
        base['protected'] = self._protected
        if self._user_data is not None:
            base['user_data'] = self.user_data
        return base

    def lock(self):
        """The lock() method will also lock the pure gases."""
        self._locked = True
        for gas in self.pure_gases:
            gas.lock()

    def unlock(self):
        """The unlock() method will also unlock the pure gases."""
        self._locked = False
        for gas in self.pure_gases:
            gas.unlock()

    @staticmethod
    def extract_all_from_xml_file(xml_file):
        """Extract all Gas objects from a THERM XML file.

        Args:
            xml_file: A path to an XML file containing objects Gas objects and
                corresponding PureGas components.

        Returns:
            A tuple with two elements

            -   gases: A list of all Gas objects in the XML file as fairyfly_therm
                Gas objects.

            -   pure_gases: A list of all PureGas objects in the XML file as
                fairyfly_therm PureGas objects.
        """
        # read the file and get the root
        tree = ET.parse(xml_file)
        root = tree.getroot()
        # extract all of the PureGas objects
        pure_dict = {}
        for gas_obj in root:
            if gas_obj.tag == 'PureGas':
                try:
                    xml_name = gas_obj.find('Name')
                    pure_dict[xml_name.text] = PureGas.from_therm_xml(gas_obj)
                except Exception:  # not a valid pure gas material
                    pass
        # extract all of the gas objects
        gases = []
        for gas_obj in root:
            if gas_obj.tag == 'Gas':
                try:
                    gases.append(Gas.from_therm_xml(gas_obj, pure_dict))
                except Exception:  # not a valid gas material
                    pass
        return gases, list(pure_dict.values())

    def _weighted_avg_coeff_property(self, attr, t_kelvin):
        """Get a weighted average property given a dictionary of coefficients."""
        property = []
        for gas in self._pure_gases:
            property.append(
                getattr(gas, attr + '_a') +
                getattr(gas, attr + '_b') * t_kelvin +
                getattr(gas, attr + '_c') * t_kelvin ** 2)
        return sum(tuple(pr * frac for pr, frac in zip(property, self._gas_fractions)))

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (self.identifier, self.gas_fractions) + \
            tuple(hash(g) for g in self.pure_gases)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Gas) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return 'THERM Gas: {}'.format(self.display_name)

    def __copy__(self):
        new_obj = Gas(
            [g.duplicate() for g in self.pure_gases],
            self.gas_fractions, self.identifier)
        new_obj._display_name = self._display_name
        new_obj._protected = self._protected
        new_obj._user_data = None if self._user_data is None else self._user_data.copy()
        return new_obj



# coding=utf-8
"""Boundary Therm Properties."""
from fairyfly.checkdup import is_equivalent

from ..condition.steadystate import SteadyState
from ..lib.conditions import exterior


class BoundaryThermProperties(object):
    """Therm Properties for Fairyfly Boundary.

    Args:
        host: A fairyfly_core Boundary object that hosts these properties.
        condition: An optional Condition object to set the conductive properties
            of the Boundary. The default is set to a generic concrete condition.
        u_factor_tag: An optional text string for to define a U-Factor tag along
            the boundary condition. This tag is used tell THERM the boundary on
            which a net U-Value shall be computed. Typical values to input here,
            which are recognizable in LBNL WINDOW include the following.

            * Frame
            * Edge
            * Spacer
            * ShadeInETag
            * ShadeOutETag
            * SHGC Exterior
            * New

    Properties:
        * host
        * condition
        * u_factor_tag
    """
    __slots__ = ('_host', '_condition', '_u_factor_tag')

    def __init__(self, host, condition=None, u_factor_tag=None):
        """Initialize Boundary THERM properties."""
        self._host = host
        self.condition = condition
        self.u_factor_tag = u_factor_tag

    @property
    def host(self):
        """Get the Boundary object hosting these properties."""
        return self._host

    @property
    def condition(self):
        """Get or set a THERM Condition for the boundary."""
        if self._condition:  # set by user
            return self._condition
        return exterior

    @condition.setter
    def condition(self, value):
        if value is not None:
            assert isinstance(value, SteadyState), \
                'Expected SteadyState. Got {}.'.format(type(value))
            value.lock()  # lock editing in case condition has multiple references
        self._condition = value

    @property
    def u_factor_tag(self):
        """Get or set a string for to define a U-Factor tag along the boundary condition.
        """
        return self._u_factor_tag

    @u_factor_tag.setter
    def u_factor_tag(self, value):
        if value is not None:
            try:
                value = str(value)
            except UnicodeEncodeError:  # Python 2 machine lacking the character set
                pass  # keep it as unicode
        self._u_factor_tag = value

    @classmethod
    def from_dict(cls, data, host):
        """Create BoundaryThermProperties from a dictionary.

        Note that the dictionary must be a non-abridged version for this
        classmethod to work.

        Args:
            data: A dictionary representation of BoundaryThermProperties with the
                format below.
            host: A Boundary object that hosts these properties.

        .. code-block:: python

            {
            "type": 'BoundaryThermProperties',
            "condition": {},  # A SteadyState dictionary
            "u_factor_tag": "Frame"  # text for the u-factor tag
            }
        """
        assert data['type'] == 'BoundaryThermProperties', \
            'Expected BoundaryThermProperties. Got {}.'.format(data['type'])
        new_prop = cls(host)
        if 'condition' in data and data['condition'] is not None:
            new_prop.condition = SteadyState.from_dict(data['condition'])
        if 'u_factor_tag' in data and data['u_factor_tag'] is not None:
            new_prop.u_factor_tag = data['u_factor_tag']
        return new_prop

    def apply_properties_from_dict(self, abridged_data, conditions):
        """Apply properties from a BoundaryThermPropertiesAbridged dictionary.

        Args:
            abridged_data: A BoundaryThermPropertiesAbridged dictionary (typically
                coming from a Model).
            conditions: A dictionary of conditions with condition identifiers
                as keys, which will be used to re-assign conditions.
        """
        if 'condition' in abridged_data and abridged_data['condition'] is not None:
            try:
                self.condition = conditions[abridged_data['condition']]
            except KeyError:
                raise ValueError('Boundary condition "{}" was not found in '
                                 'conditions.'.format(abridged_data['condition']))
        if 'u_factor_tag' in abridged_data and abridged_data['u_factor_tag'] is not None:
            self.u_factor_tag = abridged_data['u_factor_tag']

    def to_dict(self, abridged=False):
        """Return therm properties as a dictionary.

        Args:
            abridged: Boolean to note whether the full dictionary describing the
                object should be returned (False) or just an abridged version (True).
                Default: False.
        """
        base = {'therm': {}}
        base['therm']['type'] = 'BoundaryThermProperties' if not \
            abridged else 'BoundaryThermPropertiesAbridged'
        if self._condition is not None:
            base['therm']['condition'] = \
                self._condition.identifier if abridged else self._condition.to_dict()
        if self.u_factor_tag is not None:
            base['therm']['u_factor_tag'] = self.u_factor_tag
        return base

    def duplicate(self, new_host=None):
        """Get a copy of this object.

        Args:
            new_host: A new Boundary object that hosts these properties.
                If None, the properties will be duplicated with the same host.
        """
        _host = new_host or self._host
        return BoundaryThermProperties(_host, self._condition, self._u_factor_tag)

    def is_equivalent(self, other):
        """Check to see if these therm properties are equivalent to another object.

        This will only be True if all properties match (except for the host) and
        will otherwise be False.
        """
        if not is_equivalent(self._condition, other._condition):
            return False
        if self._u_factor_tag != other._u_factor_tag:
            return False
        return True

    def ToString(self):
        return self.__repr__()

    def __repr__(self):
        return 'Boundary Therm Properties: [host: {}]'.format(self.host.display_name)



# coding=utf-8
"""Methods to write Fairyfly core objects to THERM XML and THMZ."""
import os
import uuid
import random
import math
import datetime
import zipfile
import json
import tempfile
import xml.etree.ElementTree as ET

from ladybug_geometry.geometry2d import Polygon2D
from ladybug_geometry.geometry3d import Vector3D, Point3D, LineSegment3D, Plane, \
    Polyline3D, Face3D, Polyface3D
from ladybug_geometry.bounding import bounding_box
from fairyfly.typing import clean_string, therm_id_from_uuid
from fairyfly.shape import Shape
from fairyfly.boundary import Boundary

from fairyfly_therm.config import folders
from fairyfly_therm.material import CavityMaterial
from fairyfly_therm.simulation.parameter import SimulationParameter
from fairyfly_therm.lib.conditions import adiabatic, frame_cavity

HANDLE_COUNTER = 1  # counter used to generate unique handles when necessary


def shape_to_therm_xml(shape, plane=None, polygons_element=None, reset_counter=True):
    """Generate an THERM XML Polygon Element object from a fairyfly Shape.

    Args:
        shape: A fairyfly Shape for which an THERM XML Polygon Element object will
            be returned.
        plane: An optional ladybug-geometry Plane to set the 2D coordinate system
            into which the 3D Shape will be projected to THERM space. If None
            the Face3D.plane of the Shape's geometry will be used. (Default: None).
        polygons_element: An optional XML Element for the Polygons to which the
            generated Element will be added. If None, a new XML Element
            will be generated. (Default: None).
        reset_counter: A boolean to note whether the global counter for unique
            handles should be reset after the method is run. (Default: True).

    .. code-block:: xml

        <Polygon>
            <UUID>9320589a-2ee0-bab0-72c3f49441f3</UUID>
            <ID>1</ID>
            <MaterialUUID>8dd145d0-5f30-11ea-bc55-0242ac130003</MaterialUUID>
            <MaterialName>Laminated panel</MaterialName>
            <Origin>
                <x>0</x>
                <y>0</y>
            </Origin>
            <Points>
                <Point>
                    <x>181</x>
                    <y>-219</y>
                </Point>
                <Point>
                    <x>181</x>
                    <y>-371.4</y>
                </Point>
                <Point>
                    <x>200</x>
                    <y>-371.4</y>
                </Point>
                <Point>
                    <x>200</x>
                    <y>-219</y>
                </Point>
            </Points>
            <Type>Material</Type>
        </Polygon>
    """
    global HANDLE_COUNTER  # declare that we will edit the global variable
    # create a new Polygon element if one is not specified
    if polygons_element is not None:
        xml_poly = ET.SubElement(polygons_element, 'Polygon')
    else:
        xml_poly = ET.Element('Polygon')
    # add all of the required basic attributes
    xml_uuid = ET.SubElement(xml_poly, 'UUID')
    xml_uuid.text = shape.therm_uuid
    xml_id = ET.SubElement(xml_poly, 'ID')
    xml_id.text = str(HANDLE_COUNTER)
    HANDLE_COUNTER += 1
    xml_mat_id = ET.SubElement(xml_poly, 'MaterialUUID')
    shape_mat = shape.properties.therm.material
    xml_mat_id.text = shape_mat.therm_uuid \
        if isinstance(shape_mat, CavityMaterial) else shape_mat.identifier
    xml_mat_name = ET.SubElement(xml_poly, 'MaterialName')
    xml_mat_name.text = shape.properties.therm.material.display_name
    # add an origin
    xml_origin = ET.SubElement(xml_poly, 'Origin')
    for coord in ('x', 'y'):
        xml_oc = ET.SubElement(xml_origin, coord)
        xml_oc.text = '0'
    # add all of the geometry
    xml_points = ET.SubElement(xml_poly, 'Points')
    polygon = shape.geometry.polygon2d.vertices if plane is None else \
        [plane.xyz_to_xy(pt3) for pt3 in shape.geometry.vertices]
    for pt_2d in polygon:
        xml_point = ET.SubElement(xml_points, 'Point')
        xml_x = ET.SubElement(xml_point, 'x')
        xml_x.text = str(round(pt_2d.x, 1))
        xml_y = ET.SubElement(xml_point, 'y')
        xml_y.text = str(round(pt_2d.y, 1))
    # add the cavity ID if it exists
    if shape.user_data is not None and 'cavity_id' in shape.user_data:
        xml_cav_id = ET.SubElement(xml_poly, 'CavityUUID')
        xml_cav_id.text = str(shape.user_data['cavity_id'])
    # add the type of polygon
    xml_type = ET.SubElement(xml_poly, 'Type')
    xml_type.text = 'Material'
    if reset_counter:  # reset the counter back to 1 if requested
        HANDLE_COUNTER = 1
    return xml_poly


def boundary_to_therm_xml(boundary, plane=None, boundaries_element=None,
                          reset_counter=True):
    """Generate an THERM XML Boundary Element object from a fairyfly Boundary.

    Args:
        boundary: A fairyfly Boundary for which an THERM XML Boundary Element
            object will be returned.
        plane: An optional ladybug-geometry Plane to set the 2D coordinate
            system into which the 3D Boundary will be projected to THERM space.
            If None, it will be assumed that the Boundary lies in the World XY
            plane. (Default: None).
        boundaries_element: An optional XML Element for the Boundaries to which the
            generated objects will be added. If None, a new XML Element
            will be generated. (Default: None).
        reset_counter: A boolean to note whether the global counter for unique
            handles should be reset after the method is run. (Default: True).

    .. code-block:: xml

        <Boundary>
            <ID>45</ID>
            <UUID>14264c7e-1801-a3c1-0e115d8227ac</UUID>
            <Name>NFRC 100-2010 Exterior</Name>
            <FluxTag></FluxTag>
            <IsBlocking>true</IsBlocking>
            <NeighborPolygonUUID>5b9e5933-1080-4e9e-5c3b537d8230</NeighborPolygonUUID>
            <Origin>
                <x>0</x>
                <y>0</y>
            </Origin>
            <StartPoint>
                <x>235.670456</x>
                <y>-147.081726</y>
            </StartPoint>
            <EndPoint>
                <x>235.670456</x>
                <y>-297.081238</y>
            </EndPoint>
            <Side>0</Side>
            <ThermalEmissionProperties>
                <Emissivity>0.84</Emissivity>
                <Temperature>0</Temperature>
                <UseGlobalEmissivity>true</UseGlobalEmissivity>
            </ThermalEmissionProperties>
            <IsIlluminated>false</IsIlluminated>
            <EdgeID>0</EdgeID>
            <Type>Boundary Condition</Type>
            <Color>0x000000</Color>
            <Status>0</Status>
        </Boundary>
    """
    global HANDLE_COUNTER  # declare that we will edit the global variable
    # create a new Boundaries element if one is not specified
    if boundaries_element is None:
        boundaries_element = ET.Element('Boundaries')
    # determine an edge ID and color to be used for all segments in the boundary
    edge_id = str(random.randint(10000000, 99999999))
    color = boundary.properties.therm.condition.color.to_hex().replace('#', '0x')
    # loop through each of the line segments and add a Boundary element
    for i, seg in enumerate(boundary.geometry):
        # add all of the required basic attributes
        xml_bound = ET.SubElement(boundaries_element, 'Boundary')
        xml_id = ET.SubElement(xml_bound, 'ID')
        xml_id.text = str(HANDLE_COUNTER)
        HANDLE_COUNTER += 1
        xml_uuid = ET.SubElement(xml_bound, 'UUID')
        xml_uuid.text = boundary.therm_uuid[:-12] + str(uuid.uuid4())[-12:]
        xml_name = ET.SubElement(xml_bound, 'Name')
        xml_name.text = boundary.properties.therm.condition.display_name
        xml_flux = ET.SubElement(xml_bound, 'FluxTag')
        if boundary.properties.therm.u_factor_tag is not None:
            xml_flux.text = boundary.properties.therm.u_factor_tag
        xml_blocks = ET.SubElement(xml_bound, 'IsBlocking')
        xml_blocks.text = 'true'
        # add the UUIDs of the neighboring shapes
        if boundary.user_data is not None and 'adj_polys' in boundary.user_data:
            adj_ids = boundary.user_data['adj_polys'][i]
            for j, adj_id in enumerate(adj_ids):
                if j == 0:
                    xml_ajd_p = ET.SubElement(xml_bound, 'NeighborPolygonUUID')
                else:
                    xml_ajd_p = ET.SubElement(
                        xml_bound, 'NeighborPolygonUUID{}'.format(j + 1))
                xml_ajd_p.text = adj_id
        # add an origin
        xml_origin = ET.SubElement(xml_bound, 'Origin')
        for coord in ('x', 'y'):
            xml_oc = ET.SubElement(xml_origin, coord)
            xml_oc.text = '0'
        # add the boundary geometry
        pts_2d = seg.vertices if plane is None else \
            [plane.xyz_to_xy(pt3) for pt3 in seg.vertices]
        for k, pt_2d in enumerate(pts_2d):
            xml_point = ET.SubElement(xml_bound, 'StartPoint') if k == 0 else \
                ET.SubElement(xml_bound, 'EndPoint')
            xml_x = ET.SubElement(xml_point, 'x')
            xml_x.text = str(round(pt_2d.x, 1))
            xml_y = ET.SubElement(xml_point, 'y')
            xml_y.text = str(round(pt_2d.y, 1))
        # add the various thermal properties
        xml_side = ET.SubElement(xml_bound, 'Side')
        xml_side.text = '0'
        xml_e_prop = ET.SubElement(xml_bound, 'ThermalEmissionProperties')
        xml_emiss = ET.SubElement(xml_e_prop, 'Emissivity')
        if boundary.user_data is not None and 'emissivities' in boundary.user_data:
            xml_emiss.text = str(boundary.user_data['emissivities'][i])
        else:
            xml_emiss.text = '0.9'
        xml_temp = ET.SubElement(xml_e_prop, 'Temperature')
        xml_temp.text = '0'
        xml_g_emiss = ET.SubElement(xml_e_prop, 'UseGlobalEmissivity')
        xml_g_emiss.text = 'true'
        xml_is_ill = ET.SubElement(xml_bound, 'IsIlluminated')
        xml_is_ill.text = 'false'
        # add the final identifying properties
        xml_edge_id = ET.SubElement(xml_bound, 'EdgeID')
        xml_edge_id.text = edge_id
        if boundary.user_data is not None and 'enclosure_numbers' in boundary.user_data:
            xml_enclosure = ET.SubElement(xml_bound, 'EnclosureNumber')
            xml_enclosure.text = boundary.user_data['enclosure_numbers'][i]
            xml_type = ET.SubElement(xml_bound, 'Type')
            xml_type.text = 'Frame Cavity'
        else:
            xml_type = ET.SubElement(xml_bound, 'Type')
            xml_type.text = 'Boundary Condition'
        xml_color = ET.SubElement(xml_bound, 'Color')
        xml_color.text = color
        xml_status = ET.SubElement(xml_bound, 'Status')
        if boundary.user_data is not None and 'enclosure_numbers' in boundary.user_data:
            xml_status.text = '64'
        else:
            xml_status.text = '0'
    if reset_counter:  # reset the counter back to 1 if requested
        HANDLE_COUNTER = 1
    return boundaries_element


def _cavity_to_therm_xml(properties, cavities_element=None):
    """Generate an THERM XML Cavity Element object from a list of properties.

    Args:
        properties: A list of cavity properties, including emissivities and cavity area.
        cavities_element: An optional XML Element for the Cavities to which the
            generated objects will be added. If None, a new XML Element
            will be generated. (Default: None).

    .. code-block:: xml

        <Cavity>
            <UUID>adfca13b-af61-a0d7-d29f0fbefbcc</UUID>
            <HeatFlowDirection>Unknown</HeatFlowDirection>
            <Emissivity1>0.9</Emissivity1>
            <Emissivity2>0.9</Emissivity2>
            <Temperature1>7</Temperature1>
            <Temperature2>-4</Temperature2>
            <MaxXDimension>-1</MaxXDimension>
            <MaxYDimension>-1</MaxYDimension>
            <ActualHeight>1000</ActualHeight>
            <Area>3.260304e-12</Area>
            <LocalEmissivities>false</LocalEmissivities>
            <Pressure>1.013e+05</Pressure>
            <WarmLocator>
                <x>0</x>
                <y>0</y>
            </WarmLocator>
            <ColdLocator>
                <x>0</x>
                <y>0</y>
            </ColdLocator>
        </Cavity>
    """
    # create a new Cavity element if one is not specified
    if cavities_element is not None:
        xml_cav = ET.SubElement(cavities_element, 'Cavity')
    else:
        xml_cav = ET.Element('Cavity')
    # add all of the required basic attributes
    xml_uuid = ET.SubElement(xml_cav, 'UUID')
    xml_uuid.text = str(properties[0])
    xml_hfd = ET.SubElement(xml_cav, 'HeatFlowDirection')
    xml_hfd.text = 'Unknown'
    xml_e1 = ET.SubElement(xml_cav, 'Emissivity1')
    xml_e1.text = str(properties[1])
    xml_e2 = ET.SubElement(xml_cav, 'Emissivity2')
    xml_e2.text = str(properties[2])
    xml_t1 = ET.SubElement(xml_cav, 'Temperature1')
    xml_t1.text = '7'
    xml_t2 = ET.SubElement(xml_cav, 'Temperature2')
    xml_t2.text = '-4'
    xml_mx = ET.SubElement(xml_cav, 'MaxXDimension')
    xml_mx.text = '-1'
    xml_my = ET.SubElement(xml_cav, 'MaxYDimension')
    xml_my.text = '-1'
    xml_ah = ET.SubElement(xml_cav, 'ActualHeight')
    xml_ah.text = '1000'
    xml_ar = ET.SubElement(xml_cav, 'Area')
    xml_ar.text = str(properties[3])
    xml_le = ET.SubElement(xml_cav, 'LocalEmissivities')
    xml_le.text = 'true'
    xml_pressure = ET.SubElement(xml_cav, 'Pressure')
    xml_pressure.text = '1.01e+05'
    # add a warm locator
    xml_warm = ET.SubElement(xml_cav, 'WarmLocator')
    for coord in ('x', 'y'):
        xml_oc = ET.SubElement(xml_warm, coord)
        xml_oc.text = '0'
    # add a cold locator
    xml_cold = ET.SubElement(xml_cav, 'ColdLocator')
    for coord in ('x', 'y'):
        xml_oc = ET.SubElement(xml_cold, coord)
        xml_oc.text = '0'
    return xml_cav


def model_to_therm_xml(model, simulation_par=None):
    """Generate an THERM XML Element object for a fairyfly Model.

    The resulting Element has all geometry (Shapes and Boundaries).

    Args:
        model: A fairyfly Model for which a THERM XML ElementTree object will
            be returned.
        simulation_par: A fairyfly-therm SimulationParameter object to specify
            how the THERM simulation should be run. If None, default simulation
            parameters will be generated. (Default: None).
    """
    global HANDLE_COUNTER  # declare that we will edit the global variable
    # check that we have at least one shape to translate
    assert len(model.shapes) > 0, \
        'Model must have at least one Shape to translate to THERM.'
    # duplicate model to avoid mutating it as we edit it for THERM export
    original_model = model
    model = model.duplicate()
    # scale the model if the units are not millimeters
    if model.units != 'Millimeters':
        model.convert_to_units('Millimeters')
    # remove degenerate geometry within THERM native tolerance
    try:
        model.remove_duplicate_vertices(0.1)
    except ValueError:
        error = 'Failed to remove degenerate Shapes.\nYour Model units system is: {}. ' \
            'Is this correct?'.format(original_model.units)
        raise ValueError(error)

    # determine the plane and the scale to be used for all geometry translation
    ang_tol = model.angle_tolerance
    min_pt, max_pt = bounding_box([s.geometry for s in model.shapes])
    origin = Point3D(min_pt.x, max_pt.y, max_pt.z)
    normal = model.shapes[0].geometry.normal
    up_vec = math.degrees(Vector3D(0, 0, 1).angle(normal))
    if up_vec < ang_tol or up_vec > 180 - ang_tol:
        bp = Plane(o=origin)
    else:
        if normal.z < 0:
            normal = normal.reverse()
        bp = Plane(n=normal, o=origin)
    t_vec = (bp.x * -100) + (bp.y * 100)
    offset_origin = origin.move(t_vec)
    plane = Plane(n=bp.n, o=offset_origin)
    max_dim = max((max_pt.x - min_pt.x, max_pt.y - min_pt.y, max_pt.z - min_pt.z))
    scale = 1.0 if max_dim < 100 else 100 / max_dim

    # check that all geometries lie within the tolerance of the plane
    for shape in model.shapes:
        for pt in shape.vertices:
            if plane.distance_to_point(pt) > 0.1:
                msg = 'Not all of the model shapes lie in the same plane as ' \
                    'each other. Shape "{}" is out of plane by {} ' \
                    'millimeters.'.format(shape.full_id, plane.distance_to_point(pt))
                raise ValueError(msg)
    for bound in model.boundaries:
        for pt in bound.vertices:
            if plane.distance_to_point(pt) > 0.1:
                msg = 'Not all of the model boundaries lie in the same plane as ' \
                    'the shapes. Boundary "{}" is out of plane by {} ' \
                    'millimeters.'.format(bound.full_id, plane.distance_to_point(pt))
                raise ValueError(msg)

    # split any shapes that have holes in them
    split_shapes = []
    for shape in model.shapes:
        if shape.geometry.has_holes:
            for geo in shape.geometry.split_through_holes():
                new_shp = shape.duplicate()
                new_shp._geometry = geo
                new_shp.identifier = str(uuid.uuid4())
                split_shapes.append(new_shp)
        else:
            split_shapes.append(shape)
    if len(model.shapes) != split_shapes:
        model.shapes = split_shapes

    # ensure that all shapes are counterclockwise
    for shape in model.shapes:
        poly = Polygon2D(plane.xyz_to_xy(pt) for pt in shape.geometry)
        if poly.is_clockwise:
            shape._geometry = shape.geometry.flip()

    # intersect the shape geometries with one another
    Shape.intersect_adjacency(model.shapes, 0.1, plane)

    # determine if there are any Boundary points that do not share a Shape vertex
    boundary_pts = []
    for bound in model.boundaries:
        for seg in bound.geometry:
            for pt in seg.vertices:
                for o_pt in boundary_pts:
                    if pt.is_equivalent(o_pt, tolerance=0.1):
                        break
                else:  # the point is unique
                    boundary_pts.append(pt)
    orphaned_points = []
    for bpt in boundary_pts:
        matched = False
        for shape in model.shapes:
            if matched:
                break
            for spt in shape.vertices:
                if bpt.is_equivalent(spt, tolerance=0.1):
                    matched = True
                    break
        if not matched:  # a boundary point with no Shape
            orphaned_points.append(bpt)

    # insert extra vertices to the shapes if they do not align with boundary end points
    for or_pt in orphaned_points:
        for shape in model.shapes:
            shape.insert_vertex(or_pt, tolerance=0.1)

    # ensure that there is only one contiguous shape without holes
    shape_geos = [shape.geometry for shape in model.shapes]
    polyface = Polyface3D.from_faces(shape_geos, tolerance=0.1)
    outer_edges = polyface.naked_edges
    joined_boundary = Polyline3D.join_segments(outer_edges, tolerance=0.1)
    if len(joined_boundary) != 1:
        b_msg = 'The Shapes of the input model do not form a contiguous region ' \
            'without any holes.'
        join_faces = [Face3D(poly.vertices) for poly in joined_boundary]
        merged_faces = Face3D.merge_faces_to_holes(join_faces, 0.1)
        region_count = len(merged_faces)
        plural = 's' if region_count != 1 else ''
        hole_count = 0
        for mf in merged_faces:
            if mf.has_holes:
                hole_count += len(mf.holes)
        d_msg = '{} distinct region{} with {} total holes were found.'.format(
            region_count, plural, hole_count)
        raise ValueError('{}\n{}'.format(b_msg, d_msg))

    # gather all of the extra edges to be written as adiabatic
    adiabatic_geo, adiabatic_adj = [], []
    for edge in outer_edges:
        bnd_matched = False
        for bound in model.boundaries:
            for si, seg in enumerate(bound.geometry):
                if seg.distance_to_point(edge.p1) < 0.1 and \
                        seg.distance_to_point(edge.p2) < 0.1:
                    if seg.length - edge.length > 0.1:  # split the boundary geo
                        cpt_1 = seg.closest_point(edge.p1)
                        cpt_2 = seg.closest_point(edge.p2)
                        all_pts = [seg.p1, cpt_1, cpt_2, seg.p2]
                        pt_dists = [seg.p1.distance_to_point(pt) for pt in all_pts]
                        s_pts = [p for _, p in sorted(zip(pt_dists, all_pts),
                                                      key=lambda pair: pair[0])]
                        new_geo = list(bound.geometry)
                        new_geo.pop(si)
                        for p in range(3):
                            if s_pts[p].distance_to_point(s_pts[p + 1]) > 0.1:
                                nl = LineSegment3D.from_end_points(s_pts[p], s_pts[p + 1])
                                new_geo.append(nl)
                        bound._geometry = tuple(new_geo)
                    bnd_matched = True
                    break
            if bnd_matched:
                break
        else:  # adiabatic segment to be added at the end
            shape_matched = False
            for shape in model.shapes:
                for seg in shape.geometry.segments:
                    if edge.p1.is_equivalent(seg.p1, 0.1) or \
                            edge.p1.is_equivalent(seg.p2, 0.1):
                        if edge.p2.is_equivalent(seg.p1, 0.1) or \
                                edge.p2.is_equivalent(seg.p2, 0.1):
                            adiabatic_adj.append([shape.therm_uuid])
                            if edge.p1.is_equivalent(seg.p2, 0.1):
                                edge = edge.flip()
                            shape_matched = True
                            break
                if shape_matched:
                    break
            adiabatic_geo.append(edge)
            if not shape_matched:
                adiabatic_adj.append([])

    # gather any edges to be written with a frame cavity boundary
    frame_cavity_geo, cavity_props, enclosure_numbers = [], [], []
    enclosure_count = 1
    solid_shapes, cavity_shapes = [], []
    for shape in model.shapes:
        c_mat = shape.properties.therm.material
        if isinstance(c_mat, CavityMaterial) and c_mat.cavity_model != 'CEN':
            cavity_shapes.append(shape)
            cav_id = therm_id_from_uuid(str(uuid.uuid4()))
            cav_number = str(enclosure_count)
            enclosure_count += 1
            if shape.user_data is None:
                shape.user_data = {'cavity_id': cav_id}
            else:
                shape.user_data['cavity_id'] = cav_id
            cavity_prop = [cav_id, c_mat.emissivity, c_mat.emissivity_back,
                           shape.area * 1e-6]
            cavity_props.append(cavity_prop)
            for edge in shape.geometry.segments:
                frame_cavity_geo.append(edge)
                enclosure_numbers.append(cav_number)
        else:
            solid_shapes.append(shape)
    if len(frame_cavity_geo) == 0:
        cavity_boundary = None
        all_boundaries = model.boundaries
    else:
        cavity_boundary = Boundary(frame_cavity_geo)
        cavity_boundary.properties.therm.condition = frame_cavity
        cavity_boundary.user_data = {'enclosure_numbers': enclosure_numbers}
        all_boundaries = model.boundaries + (cavity_boundary,)

    # add the UUIDs of the polygons next to the edges to the Boundary.user_data
    ordered_shapes = cavity_shapes + solid_shapes
    for bound in all_boundaries:
        oriented_geo, bound_adj_shapes, bound_emissivity = [], [], []
        for edge in bound.geometry:
            adj_shapes, bnd_e = [], 0.9
            for shape in ordered_shapes:
                for seg in shape.geometry.segments:
                    if edge.p1.is_equivalent(seg.p1, 0.1) or \
                            edge.p1.is_equivalent(seg.p2, 0.1):
                        if edge.p2.is_equivalent(seg.p1, 0.1) or \
                                edge.p2.is_equivalent(seg.p2, 0.1):
                            adj_shapes.append(shape.therm_uuid)
                            if edge.p1.is_equivalent(seg.p2, 0.1):
                                edge = edge.flip()
                            shape_mat = shape.properties.therm.material
                            if not isinstance(shape_mat, CavityMaterial):
                                bnd_e = shape_mat.emissivity
                            break
            bound_adj_shapes.append(adj_shapes)
            bound_emissivity.append(bnd_e)
            oriented_geo.append(edge)
        bound._geometry = tuple(oriented_geo)
        if bound.user_data is None:
            bound.user_data = {
                'adj_polys': bound_adj_shapes,
                'emissivities': bound_emissivity
            }
        else:
            bound.user_data['adj_polys'] = bound_adj_shapes
            bound.user_data['emissivities'] = bound_emissivity

    # load up the template XML file for the model
    package_dir = os.path.dirname(os.path.abspath(__file__))
    template_file = os.path.join(package_dir, '_templates', 'Default.xml')
    xml_tree = ET.parse(template_file)
    xml_root = xml_tree.getroot()
    model_name = clean_string(model.display_name)

    # assign the property for the scale so it looks good in THERM
    xml_preferences = xml_root.find('Preferences')
    xml_settings = xml_preferences.find('Settings')
    xml_scale = xml_settings.find('Scale')
    xml_scale.text = str(scale)

    # set the properties for the document
    xml_props = xml_root.find('Properties')
    xml_gen = xml_props.find('General')
    therm_ver = '.'.join(str(i) for i in folders.THERM_VERSION)
    therm_ver = 'Version {}'.format(therm_ver)
    xml_calc_ver = xml_gen.find('CalculationVersion')
    xml_calc_ver.text = therm_ver
    xml_cre_ver = xml_gen.find('CreationVersion')
    xml_cre_ver.text = therm_ver
    xml_mod_ver = xml_gen.find('LastModifiedVersion')
    xml_mod_ver.text = therm_ver
    xml_cre_date = xml_gen.find('CreationDate')
    xml_cre_date.text = str(datetime.datetime.now().replace(microsecond=0))
    xml_cre_date = xml_gen.find('LastModified')
    xml_cre_date.text = str(datetime.datetime.now().replace(microsecond=0))
    xml_model_name = xml_gen.find('Title')
    xml_model_name.text = model_name

    # write the 3D plane into the notes section
    xml_notes = xml_gen.find('Notes')
    xml_notes.text = 'Plane: {}'.format(json.dumps(plane.to_dict()))

    # add the calculation options
    sim_par = simulation_par if simulation_par is not None else SimulationParameter()
    xml_calc_opt = xml_props.find('CalculationOptions')
    sim_par.mesh.to_therm_xml(xml_calc_opt)

    # write all of the cavity definitions into the model
    if len(cavity_props) != 0:
        xml_cavities = ET.SubElement(xml_root, 'Cavities')
        for cp in cavity_props:
            _cavity_to_therm_xml(cp, xml_cavities)

    # translate all Shapes to polygons
    xml_polygons = ET.SubElement(xml_root, 'Polygons')
    for shape in model.shapes:
        shape_to_therm_xml(shape, plane, xml_polygons, reset_counter=False)

    # translate all Boundaries
    xml_boundaries = ET.SubElement(xml_root, 'Boundaries')
    for bound in model.boundaries:
        # remove any boundary geometries that are not assigned to shapes
        adj_polys, seg_es = bound.user_data['adj_polys'], bound.user_data['emissivities']
        new_segs, new_adj_polys, new_seg_es = [], [], []
        for seg, adj_poly, seg_e in zip(bound.geometry, adj_polys, seg_es):
            if len(adj_poly) != 0:
                new_segs.append(seg)
                new_adj_polys.append(adj_poly)
                new_seg_es.append(seg_e)
        # write the boundary into the XML
        if len(new_segs) != 0:
            bound._geometry = tuple(new_segs)
            bound.user_data['adj_polys'] = new_adj_polys
            bound.user_data['emissivities'] = new_seg_es
            boundary_to_therm_xml(bound, plane, xml_boundaries, reset_counter=False)

    # add the extra adiabatic boundaries
    ad_bnd = Boundary(adiabatic_geo)
    ad_bnd.properties.therm.condition = adiabatic
    ad_bnd.user_data = {'adj_polys': adiabatic_adj}
    boundary_to_therm_xml(ad_bnd, plane, xml_boundaries, reset_counter=False)

    # add the cavity boundaries if they exist
    if cavity_boundary is not None:
        boundary_to_therm_xml(cavity_boundary, plane, xml_boundaries, reset_counter=False)

    # reset the handle counter back to 1 and return the root XML element
    HANDLE_COUNTER = 1
    return xml_root


def shape_to_therm_xml_str(shape):
    """Generate an THERM XML string from a fairyfly Shape.

    Args:
        shape: A fairyfly Shape for which an THERM XML Polygon string will
            be returned.
    """
    xml_root = shape_to_therm_xml(shape)
    try:  # try to indent the XML to make it read-able
        ET.indent(xml_root)
        return ET.tostring(xml_root, encoding='unicode')
    except AttributeError:  # we are in Python 2 and no indent is available
        return ET.tostring(xml_root)


def boundary_to_therm_xml_str(boundary):
    """Generate an THERM XML string from a fairyfly Boundary.

    Args:
        shape_mesh: A fairyfly Boundary for which an THERM XML Boundary string
            will be returned.
    """
    xml_root = boundary_to_therm_xml(boundary)
    try:  # try to indent the XML to make it read-able
        ET.indent(xml_root)
        return ET.tostring(xml_root, encoding='unicode')
    except AttributeError:  # we are in Python 2 and no indent is available
        return ET.tostring(xml_root)


def model_to_therm_xml_str(model, simulation_par=None):
    """Generate a THERM XML string for a Model.

    The resulting Element has all geometry (Shapes and Boundaries).

    Args:
        model: A fairyfly Model for which an THERM XML text string will be returned.
        simulation_par: A fairyfly-therm SimulationParameter object to specify
            how the THERM simulation should be run. If None, default simulation
            parameters will be generated. (Default: None).
    """
    # create the XML string
    xml_root = model_to_therm_xml(model, simulation_par)
    try:  # try to indent the XML to make it read-able
        ET.indent(xml_root, '\t')
        return ET.tostring(xml_root, encoding='unicode')
    except AttributeError:  # we are in Python 2 and no indent is available
        return ET.tostring(xml_root)


def model_to_thmz(model, output_file, simulation_par=None):
    """Write a THERM Zip (.thmz) file from a Fairyfly Model.

    Args:
        model: A fairyfly Model for which an THMZ file will be written.
        output_file: The path to the THMZ file that will be written from the model.
        simulation_par: A fairyfly-therm SimulationParameter object to specify
            how the THERM simulation should be run. If None, default simulation
            parameters will be generated. (Default: None).

    Usage:

    .. code-block:: python

        import os
        from fairyfly.model import Model
        from fairyfly.config import folders
        from fairyfly_therm.lib.materials import concrete, air_cavity
        from fairyfly_therm.lib.conditions import exterior, interior

        # Crate an input Model
        model = Model.from_layers([100, 200, 100], height=1000)
        model.shapes[0].properties.therm.material = concrete
        model.shapes[1].properties.therm.material = air_cavity
        model.shapes[2].properties.therm.material = concrete
        model.boundaries[0].properties.therm.condition = exterior
        model.boundaries[1].properties.therm.condition = interior
        model.display_name = 'Roman Bath Wall'

        # create the THERM Zip file for the model
        thmz = os.path.join(folders.default_simulation_folder, 'model.thmz')
        xml_str = model.to_thmz(thmz)
    """
    # make sure the directory exists where the file will be written
    dir_name = os.path.dirname(os.path.abspath(output_file))
    if not os.path.isdir(dir_name):
        os.makedirs(dir_name)
    # create a temporary directory where everything will be zipped
    therm_trans_dir = tempfile.gettempdir()
    files_to_zip = []

    # prepare the files for materials and gases
    mat_file = os.path.join(therm_trans_dir, 'Materials.xml')
    gas_file = os.path.join(therm_trans_dir, 'Gases.xml')
    gases, pure_gases = [], []
    xml_materials = ET.Element('Materials')
    xml_gases = ET.Element('Gases')
    for xml_rt in (xml_materials, xml_gases):
        xml_ver = ET.SubElement(xml_rt, 'Version')
        xml_ver.text = '1'
    # rename any materials with duplicate display names
    model_mats = model.properties.therm.materials
    mat_names, reset_dict = {}, {}
    for mat in model_mats:
        if mat.display_name in mat_names:
            mat_names[mat.display_name] += 1
            reset_dict[mat.display_name] = mat
            mat.unlock()
            mat.display_name = mat.display_name + '_' + str(mat_names[mat.display_name])
            mat.lock()
        else:
            mat_names[mat.display_name] = 1
    # write the materials and gases to a file
    for mat in model_mats:
        mat.to_therm_xml(xml_materials)
        if isinstance(mat, CavityMaterial):
            gases.append(mat.gas)
            for pg in mat.gas.pure_gases:
                pure_gases.append(pg)
    _xml_element_to_file(xml_materials, mat_file)
    files_to_zip.append(mat_file)
    for pg in pure_gases:
        pg.to_therm_xml(xml_gases)
    for g in gases:
        g.to_therm_xml(xml_gases)
    _xml_element_to_file(xml_gases, gas_file)
    files_to_zip.append(gas_file)

    # write the Model into the temporary directory
    model_file = os.path.join(therm_trans_dir, 'Model.xml')
    xml_model = model_to_therm_xml(model, simulation_par)
    xml_props = xml_model.find('Properties')
    xml_gen = xml_props.find('General')
    xml_direct = xml_gen.find('Directory')
    xml_direct.text = dir_name
    xml_file_name = xml_gen.find('FileName')
    xml_file_name.text = os.path.basename(output_file.replace('.thmz', ''))
    _xml_element_to_file(xml_model, model_file)
    files_to_zip.append(model_file)

    # write the boundary conditions to a file
    bc_file = os.path.join(therm_trans_dir, 'SteadyStateBC.xml')
    xml_bcs = ET.Element('BoundaryConditions')
    xml_ver = ET.SubElement(xml_bcs, 'Version')
    xml_ver.text = '1'
    for bc in [adiabatic] + model.properties.therm.conditions:
        bc.to_therm_xml(xml_bcs)
    _xml_element_to_file(xml_bcs, bc_file)
    files_to_zip.append(bc_file)

    # put back any materials that were edited
    for mat_name, mat in reset_dict.items():
        mat.unlock()
        mat.display_name = mat_name
        mat.lock()

    # zip everything together
    with zipfile.ZipFile(output_file, 'w', zipfile.ZIP_DEFLATED) as zipf:
        for file in files_to_zip:
            # Add the file to the zip archive
            # arcname=os.path.basename(file) ensures only the filename is used
            # inside the zip, avoiding unwanted directory structures
            zipf.write(file, arcname=os.path.basename(file))

    return output_file


def _xml_element_to_file(xml_root, file_path):
    """Write an XML element to a file."""
    try:  # try to indent the XML to make it read-able
        ET.indent(xml_root, '\t')
        xml_str = ET.tostring(xml_root, encoding='unicode', short_empty_elements=False)
    except AttributeError:  # we are in Python 2 and no indent is available
        xml_str = ET.tostring(xml_root)
    with open(file_path, 'wb') as fp:
        fp.write(xml_str.encode('utf-8'))
